#ifndef kerevents_h
#define kerevents_h

#include <sys/kev.h>


#pragma pack(1);


/*-
  Events without a defined structure:
  _EVENT_KER_SPECIAL
  _EVENT_KER_APM
  _EVENT_KER_ENDFAULT
  _EVENT_KER_ERROR
  _EVENT_KER_NOP
  _EVENT_KER_SEMAPHORE
*/

/*-
  Events with no followup data:

  _EVENT_KER_ACTIVE       The active process changed.
  	- immediate is the (newly) active process.

  _EVENT_KER_READY        A process has been preempted.
    - immediate is the preempted process.

  _EVENT_KER_IRQDONE      An interrupt has completed processing.
    - immediate is the irq number

  _EVENT_USER_IRQDONE     An interrupt handler has completed.
    - immediate is the irq number.

  _EVENT_KER_DEATH        A process is dead.
    - immediate is the pid of deceased.

*/

/*-
  _EVENT_KER_PROC_STAT    A new process is created.
    - immediate is the pid of the newly born.
    [note: proc's are actually created before this message,
     this message finalizes the creation ]
  The structure _event_ker_proc_stat defines it.
*/


/*-
 _EVENT_KER_IRQ           - beginning of an interrupt
        -    immediate is PID, 0 for a nested interrupt.
 _EVENT_KER_USER_IRQ      - start a user-supplied interrupt handler.
        -    immediate is PINDEX who owns interrupt
 */
struct _event_ker_irq {
	unsigned long  ip;
	unsigned short cs;
	short          irq;
};

/*-
 _EVENT_KER_FAULT         - a user space processor fault.
	- immediate is the PID.
*/
struct _event_ker_fault {
	unsigned long  ip;
	unsigned short cs;
	unsigned short faultno;
};

/*-
*/
struct _event_ker_user_irq {
	unsigned short cs;
	short          irq;
};

/*-
 For each of these events, the immediate is the calling pid.

 _EVENT_KER_SEND     - send to a filedescriptor
               immediate is invoker
 _EVENT_KER_SENDFD   
               immediate is invoker
 _EVENT_KER_RECV
               immediate is invoker
 _EVENT_KER_REPLY
               immediate is invoker
 _EVENT_KER_CRECEIVE
               immediate is invoker
 _EVENT_KER_NETDATA  - manipulate network q.
               immediate is invoker, data is type type.
 _EVENT_KER_KILL     - send a signal to a process
               immediate is invoker, data is signal #.

*/
struct _event_ker_msg {
	unsigned short targetpid;
	unsigned char  targetstate;
	unsigned char  mystate;
	unsigned long  data;
};

/*-
 _EVENT_KER_RELAY    - redirect a sent message to another server.
               immediate is the relayer, target is the relayee.
 _EVENT_KER_PROXY    - proxy sends to a process.
               immediate is the proxy, target is the owning process.
 _EVENT_KER_TRIGGER  - process triggers a proxy.
               immediate is invoker, target is proxy.
 _EVENT_KER_READMSG  - receiver extracts data from sender.
               immediate is invoker
 _EVENT_KER_WRITEMSG - receiver sets data in sender.
               immediate is invoker
 */
struct _event_ker_nomsg {
	unsigned short  targetpid;
	unsigned char   targetstate;
	unsigned char   mystate;
};

/*-
 _EVENT_KER_SRET     - proc returns from signal handler
 _EVENT_KER_YIELD    - proc yields.
 */
struct _event_ker_sret {
	unsigned char   notused[3];
	unsigned char   mystate;
};

/*-
 _EVENT_KER_PRIO     - proc changed priority
*/

struct _event_ker_prio {
	unsigned char   priority;
	unsigned char   notused[3];
};

/*-
 _EVENT_KER_SEM      - semaphore operation
       immediate is invoker
*/

struct _event_ker_sem {
	unsigned long   semid;
	unsigned char   operation;
#define ev_semwait 0
#define ev_sempost 1
	unsigned char   notused[3];
};



typedef union {
	struct _event_ker_irq        irq;
	struct _event_ker_fault      fault;
	struct _event_ker_user_irq   uirq;
	struct _event_ker_msg        ipc0;
	struct _event_ker_nomsg      ipc1;
	struct _event_ker_sret       sret;
	struct _event_ker_prio       prio;
	struct _event_ker_sem        sem;
/*-
 * This one is a bit magic.
 * the lowest level dispatch notices the event is
 * pointer based and dereferences it
 */
	struct _event_ker_proc_stat  stat;
} EventData;

#pragma pack()

#endif
