#include <libc.h>
#include <sys/psinfo.h>
#include <sys/sigcontext.h>
#include "kevio.h"
#include "private.h"

enum IpcType {
	IpcSend,
	IpcRecv,
	IpcReply,
	IpcRead,
	IpcWrite
};

int
print_ipc(KRec *e, int type, char *src, char *dest, FILE *fp)
{
	static int init = 1;
	char          msg[20];
	char	*p;

	if (init) {
		msg_open(0);
		init = 0;
	}

	if (e->kr_nwords > 1) {
		sprintf(msg, "(0x%x:0x%x)",
			 e->kr_data[1] & 0xffff,
			 e->kr_data[1] >> 16);
	} else {
		strcpy(msg, "(none)");
	}
	switch (type) {
	case IpcSend:
		p = msg_lookup(&e->kr_data[1], (e->kr_nwords-1)*4);
		fprintf(fp, "%s sent %s [%s] to %s\n", src, msg, msg_symbolic_name(p), dest);
		break;
	case IpcRecv:
		p = msg_lookup(&e->kr_data[1], (e->kr_nwords-1)*4);
		fprintf(fp, "%s recv %s [%s] from %s\n", src, msg, msg_symbolic_name(p), dest);
		break;
	case IpcReply:
		fprintf(fp, "%s reply %s [%s] to %s\n", src, msg, strerror(*(short *)&e->kr_data[1] & 0xffff), dest);
		break;
	case IpcRead:
		fprintf(fp, "%s read from %s\n", src, dest);
		break;
	case IpcWrite:
		fprintf(fp, "%s wrote to %s\n", src, dest);
		break;
	}
	return 0;
}

print_pstat(KFILE *fp, struct _event_ker_proc_stat *ps, FILE *outf)
{
	fprintf(outf, "pid %lu[%lu] ", ps->pid, PID_INDEX(fp, ps->pid));
	if (ps->flags & _PPF_VMID) {
		fprintf(outf,"vproxy, connects %lu to %lu:%lu\n",
			ps->un.vidinfo.local_pid,
			ps->un.vidinfo.remote_nid,
			ps->un.vidinfo.remote_pid);
	} else if (ps->flags & _PPF_VID) {
		fprintf(outf,"vc -  connects %lu to %lu:%lu\n",
			ps->un.vidinfo.local_pid,
			ps->un.vidinfo.remote_nid,
			ps->un.vidinfo.remote_pid);
	} else if (ps->flags & _PPF_MID) {
		fprintf(outf, "proxy - owned by %lu\n", ps->blocked_on);
	} else {
		fprintf(outf, "process '%s'\n", ps->un.pidinfo.name);
	}
	return 0;
}

int
GetProcName(KFILE *fp, unsigned pid, char *buf, unsigned blen)
{
	Proc      *p0;

	if (pid == 0) {
		strcpy(buf, "{none}");
		return 6;
	}
	p0 = KevGetProc(fp, pid);

	buf[0] = 0;
	if (p0) {
		if (p0->lpid) {
			Proc   *p2 = KevGetProc(fp, p0->lpid);
			if (p2) {
				return sprintf(buf, "proxy(%s)",
					p2->name?p2->name:"????");
			}
		} 
		return sprintf(buf, "%s", p0->name ? p0->name : "??");
	}
	return sprintf(buf, "PID[%u] ", pid);
}

GetFaultName(unsigned id, char *buf)
{
	switch (id) {
	case TRAP_ZDIV: strcpy(buf, "Divide by Zero"); break;
	case TRAP_DEBUG:strcpy(buf, "Debug Fault"); break;
	case TRAP_NMI:  strcpy(buf, "Non Maskable Interrupt"); break;
	case TRAP_BRKPT:strcpy(buf, "Break Point"); break;
	case TRAP_OFLOW:strcpy(buf, "Numeric Overflow"); break;
	case TRAP_BOUNDS:strcpy(buf, "Bounds Check Failed"); break;
	case TRAP_BADOP: strcpy(buf, "Invalid OpCode"); break;
	case TRAP_NONDP: strcpy(buf, "Device Not Available"); break;
	case TRAP_DFAULT:strcpy(buf, "System Error-Double Fault"); break;
	case TRAP_NDPSEGV: strcpy(buf, "NDP Invalid Address"); break;
	case TRAP_BADTSS: strcpy(buf, "System Error-Bad TSS"); break;
	case TRAP_NOTPRESENT: strcpy(buf, "Segment Not Present"); break;
	case TRAP_NOSTACK: strcpy(buf, "Stack Addressing Error"); break;
	case TRAP_GPF:     strcpy(buf, "General Protection Fault"); break;
	case TRAP_PAGE:    strcpy(buf, "Page Protection Fault"); break;
	case TRAP_FPERROR: strcpy(buf, "Floating Point Error"); break;
	default:
		sprintf(buf, "Unknown Fault Code");
	}
}

print_kerevent(KFILE *fp, KRec *e, FILE *outf)
{
	char     src[256], dest[256];
	print_time(e, outf); putc(' ', outf);
	src[0] = dest[0] = 0;
	switch (e->kr_opcode) {
	case _EVENT_KER_IRQ:
		GetProcName(fp, e->kr_oprnd, src, sizeof src);
		fprintf(outf, "%s interrupted(#%d): @%x:%lx\n",
			src,
			(short)(e->kr_data[1]>>16),
			e->kr_data[1] & 0xffff,
			e->kr_data[0]);
		break;
	case _EVENT_KER_SEND:
		GetProcName(fp, e->kr_oprnd, src, sizeof src);
		GetProcName(fp, e->kr_data[0] & 0xffff, dest, sizeof dest);
		print_ipc(e, IpcSend, src, dest, outf);
		break;
	case _EVENT_KER_RECEIVE:
		GetProcName(fp, e->kr_oprnd, src, sizeof src);
		GetProcName(fp, e->kr_data[0] & 0xffff, dest, sizeof dest);
		print_ipc(e, IpcRecv, src, dest, outf);
		break;
	case _EVENT_KER_REPLY:
		GetProcName(fp, e->kr_oprnd, src, sizeof src);
		GetProcName(fp, e->kr_data[0] & 0xffff, dest, sizeof dest);
		print_ipc(e, IpcReply, src, dest, outf);
		break;
	case _EVENT_KER_CRECEIVE:
		fprintf(outf, "crecv(0x%x)\n", e->kr_opcode);
		break;

	case _EVENT_KER_READMSG:
		GetProcName(fp, e->kr_oprnd, src, sizeof src);
		GetProcName(fp, e->kr_data[0] & 0xffff, dest, sizeof dest);
		print_ipc(e, IpcRead, src, dest, outf);
		break;
	case _EVENT_KER_WRITEMSG:
		GetProcName(fp, e->kr_oprnd, src, sizeof src);
		GetProcName(fp, e->kr_data[0] & 0xffff, dest, sizeof dest);
		print_ipc(e, IpcWrite, src, dest, outf);
		break;
	case _EVENT_KER_RELAY:
		fprintf(outf, "relay(0x%x)\n", e->kr_oprnd);
		break;
	case _EVENT_KER_SRET:
		GetProcName(fp, e->kr_oprnd, src, sizeof src);
		fprintf(outf, "%s returns from signal handler\n", src);
		break;
	case _EVENT_KER_PRIORITY:
		GetProcName(fp, e->kr_oprnd, src, sizeof src);
		fprintf(outf, "priority %s = %u\n", src, e->kr_data[0]);
		break;
	case _EVENT_KER_NETDATA:
		GetProcName(fp, e->kr_oprnd, src, sizeof src);
		GetProcName(fp, e->kr_data[0] & 0xffff, dest, sizeof dest);
		fprintf(outf, "%s Netdata(%u) to %s\n", src, e->kr_data[1], dest);
		break;
	case _EVENT_KER_YIELD:
		fprintf(outf, "%s yields\n", src);
		break;
	case _EVENT_KER_SENDFD:
		fprintf(outf, "?sendfd(0x%x)\n", e->kr_oprnd);
		break;
	case _EVENT_KER_TRIGGER:
		GetProcName(fp, e->kr_oprnd, src, sizeof src);
		GetProcName(fp, e->kr_data[0] & 0xffff, dest, sizeof dest);
		fprintf(outf, "%s Triggers %s\n", src, dest);
		break;
	case _EVENT_KER_SPECIAL:
		break;
	case _EVENT_KER_SIGNAL:
		GetProcName(fp, e->kr_oprnd, src, sizeof src);
		GetProcName(fp, e->kr_data[0]>>16, dest, sizeof dest);
		fprintf(outf, "%s kill(%u) %s\n",
			src, e->kr_data[0]&0xffff, dest);
		break;
	case _EVENT_KER_APM:
		fprintf(outf, "event [0x%x]\n", e->kr_opcode);
		break;
	case _EVENT_KER_SEMAPHORE:
		GetProcName(fp, e->kr_oprnd, src, sizeof src);
		fprintf(outf, "%s sem%s(%u)\n",
			src, e->kr_data[1] ? "post" : "wait",
			e->kr_data[0]);
		break;
	case _EVENT_KER_PROC_STAT:
		print_pstat(fp, (struct _event_ker_proc_stat *)e->kr_data, outf);
		break;
	case _EVENT_KER_DEATH:
		fprintf(outf, "%lu is dead\n", e->kr_oprnd);
		break;
	case _EVENT_KER_PROXY:
		GetProcName(fp, e->kr_oprnd, src, sizeof src);
		GetProcName(fp, e->kr_data[0] & 0xffff, dest, sizeof dest);
		fprintf(outf, "%s triggers %s\n", src, dest);
		break;
	case _EVENT_KER_ACTIVE:
		GetProcName(fp, e->kr_oprnd, src, sizeof src);
		fprintf(outf, "active is %s\n", src);
		break;
	case _EVENT_KER_READY:
		GetProcName(fp, e->kr_oprnd, src, sizeof src);
		fprintf(outf, "%s is ready\n",src);
		break;
	case _EVENT_KER_IRQDONE:
		fprintf(outf, "end of interrupt #%d\n", e->kr_oprnd);
		break;
	case _EVENT_KER_USER_IRQ:
		GetProcName(fp, e->kr_oprnd, src, sizeof src);
		fprintf(outf, "interrupt handler(%d): %s, @%04.4x:%08.5lx\n",
			(short)(e->kr_data[1]>>16),
			src,
			e->kr_data[1] & 0xffff,
			e->kr_data[0]);
		break;

	case _EVENT_KER_USER_EOI:
		if (e->kr_oprnd) {
			GetProcName(fp, e->kr_oprnd, src, sizeof src);
			fprintf(outf, "handler done, wakeup %s\n", src);
		} else {
			fprintf(outf, "handler done\n");
		}
		break;
/*- fault, */
	case _EVENT_KER_ENDFAULT:
		fprintf(outf,"end fault %u\n", e->kr_oprnd);
		break;
/*- active, eip, cs:fault#, fcode:nused */
	case _EVENT_KER_FAULT:
		GetProcName(fp, e->kr_oprnd, src, sizeof src);
		GetFaultName(e->kr_data[1]>>16, dest);
		fprintf(outf,"%s fault %s(%u) at %x:%lx\n",
			src,
			dest, e->kr_data[1]>>16,
			e->kr_data[1] & 0xffff,e->kr_data[0]);
		break;
	case _EVENT_KER_KILL:
		GetProcName(fp, e->kr_oprnd, src, sizeof src);
		GetProcName(fp, e->kr_data[0]>>16, dest, sizeof dest);
		fprintf(outf,"%s kill(%u) %s\n", src, e->kr_data[1], dest);
		break;
	case _EVENT_KER_NOP: {
			unsigned i;
			fprintf(outf, "NOP [0x%x + 0x%x]: ", e->kr_opcode, e->kr_oprnd);
			for (i=0; i < e->kr_nwords; i++) {
				fprintf(outf, "0x%x ", e->kr_data[i]);
			}
			fprintf(outf, "\n");
		}
		break;
	case _EVENT_KER_ERROR:
		GetProcName(fp, e->kr_oprnd, src, sizeof src);
		fprintf(outf, "%s kcall#%d error %u (%s)\n",
			src, e->kr_data[0] & 0xff,
			e->kr_data[0]>>16, strerror(e->kr_data[0]>>16));
		break;

	case _EVENT_USER_TRON:
		GetProcName(fp, e->kr_oprnd, src, sizeof src);
		fprintf(outf, "%s trace on %s\n", src, nlookup(e->kr_data[0]));
		break;

	case _EVENT_USER_TROFF:
		GetProcName(fp, e->kr_oprnd, src, sizeof src);
		fprintf(outf, "%s trace off %s\n", src, nlookup(e->kr_data[0]));
		break;

	case _EVENT_USER_TRFLUSH:
		GetProcName(fp, e->kr_oprnd, src, sizeof src);
		fprintf(outf, "%s trace flush\n", src);
		break;

	default:
		fprintf(outf, "??event [0x%x]\n", e->kr_opcode);
	}
	return 0;
}
