
/*
 *	topcpu.c
 *
 *	QNX 4
 *
 *	(C) Copyright 1991, 1992, 1996 by Robert Krten, all rights reserved.
 *
 *	This module is responsible for the TOP CPU display
 *
 *	1991 09 20	R. Krten		created
 *	1991 09 22	R. Krten		added histogram mode
 *	1992 01 05	R. Krten		released for free
 *	1996 11 01	R. Krten		changed "min idle" priority to 5 for screen savers under Photon
*/

#include "sysmon.h"

static	void	accumulateBackbone (int);
static	int		compareTOPCPU (TopcpuT **, TopcpuT **);
static	void	displayDeltaTOPCPU (void);
static	void	displayHistogram (void);
static	void	fetchSortTOPCPU (void);
static	void	removeDeadEntries (int);

#define	MaxHashEntries			16		/* hash is lowest four bits */
extern	char *progname;					/* main.c */
extern	char *version;					/* version.c */

extern	int	opta;
extern	int	opth;
extern	int	optims;

extern struct _osinfo osData;
struct _psinfo psData;
char *lines = "";

static	int numTopcpuLines;				/* 15 for a 25 line screen */
static	int processCount;
static	int last80 [80];
static	long currentHistogramValue;
static	long histogramTotal;
static	long histogramSamples;

static	TopcpuT	topcpuBackbone [MaxHashEntries];

void
doTOPCPU ()
{
	int		i;
	int		histogramCount;

	/*
	 *	clear out the top CPU backbone for the hash table, and initialize
	*/

	for (i = 0; i < MaxHashEntries; i++) {
		topcpuBackbone [i].next = NULL;
	}
	processCount = 0;
	for (i = 0; i < 80; i++) {
		last80 [i] = 0;
	}


	/*
	 *	set up full screen console stuff
	*/

	numTopcpuLines = 15;
	term_cur (1, 0);
	term_clear (TERM_CLS_EOS);
	term_printf (2, 21, opta, "TOP CPU Processes %s", opth ? "(Histogram)" : "");
	term_printf (3, 0, opta, "%s", "Ŀ");
	term_printf (4, 0, opta, "%s", " PID       Name       %                                                  ");
	term_printf (5, 0, opta, "%s", "Ĵ");
	if (opth) {
		if (term_state.num_rows <= 25) {
			opth = 0;
		}
	}

	/*
	 *	start it up!
	*/

	accumulateBackbone (1);
	histogramCount = 0;
	histogramTotal = 0;
	histogramSamples = 0;
	currentHistogramValue = 0;
	if (opth) {
		displayHistogram ();		/* just to get titles, etc */
	}
	histogramSamples = 0;
	for (;;) {
		delay (optims);
		accumulateBackbone (2);
		removeDeadEntries (2);
		displayDeltaTOPCPU ();

		delay (optims);
		accumulateBackbone (1);
		removeDeadEntries (1);
		displayDeltaTOPCPU ();

		if (histogramCount++ >= 24) {
			if (opth) {
				displayHistogram ();
				currentHistogramValue = 0;
			}
			histogramCount = 0;
		}

		if (dev_ischars (0) && getc (stdin) == 'q') {
			break;
		}
	}
}

static void
accumulateBackbone (pass)
int		pass;
{
	pid_t	currentPID;
	TopcpuT	*ptr;
	TopcpuT	*prev;

	currentPID = 1;
	for (;;) {
		currentPID = qnx_psinfo (PROC_PID, currentPID, &psData, 0, NULL);
		if (currentPID == -1) {
			if (errno == ESRCH || errno == EINVAL) {
				return;				/* done, go away */
			}
		}
		if ((psData.flags & (_PPF_MID | _PPF_VID)) == 0) {
			prev = &topcpuBackbone [currentPID & 0x0f];
			for (ptr = prev -> next; ptr != NULL; ptr = (prev = ptr) -> next) {
				if (ptr -> pid == currentPID) {
					ptr -> prio = psData.priority;
					ptr -> delta = ptr -> total;
					ptr -> total = psData.un.proc.times.tms_cutime;
					ptr -> total+= psData.un.proc.times.tms_cstime;
					ptr -> total+= psData.un.proc.times.tms_stime;
					ptr -> total+= psData.un.proc.times.tms_utime;
					ptr -> delta = ptr -> total - ptr -> delta;
					break;
				}
			}
			if (ptr == NULL) {
				if ((prev -> next = (TopcpuT *) malloc (TopcpuS)) == NULL) {
					fatal ("Couldn't allocate memory for new topcpu hash entry");
				}
				ptr = prev -> next;		/* get to new entry */
				ptr -> pid = currentPID;
				ptr -> prio = psData.priority;
				fixName (ptr -> name, psData.un.proc.name);
				ptr -> total = psData.un.proc.times.tms_cutime;
				ptr -> total+= psData.un.proc.times.tms_cstime;
				ptr -> total+= psData.un.proc.times.tms_stime;
				ptr -> total+= psData.un.proc.times.tms_utime;
				ptr -> delta = 0;	/* new processes don't have any time! */
				ptr -> next = NULL;
				processCount++;
			}
			ptr -> pass = pass;
		}
		currentPID++;
	}
}

static void
removeDeadEntries (pass)
int		pass;
{
	int		i;
	TopcpuT	*ptr;
	TopcpuT	*prev;

	for (i = 0; i < MaxHashEntries; i++) {
		prev = &topcpuBackbone [i];
		ptr = prev -> next;
		while (ptr != NULL) {
			if (!*ptr -> name) {
				getName (ptr -> name, ptr -> pid);
			}
			if (ptr -> pass != pass) {
				prev -> next = ptr -> next;
				free (ptr);
				processCount--;
			} else {
				prev = ptr;
			}
			ptr = prev -> next;
		}
	}
}

static int
compareTOPCPU (a, b)
TopcpuT **a;
TopcpuT	**b;
{
	if ((*a) -> delta > (*b) -> delta) {
		return (-1);
	} else if ((*a) -> delta < (*b) -> delta) {
		return (1);
	} else {
		return (0);
	}
}

static TopcpuT **processList;

static void
fetchSortTOPCPU ()
{
	int		i;
	int		p;
	TopcpuT *ptr;

	if ((processList = (TopcpuT **) malloc (sizeof (TopcpuT *) * processCount)) == NULL) {
		fatal ("Couldn't allocate memory for new processList");
	}
	p = 0;
	for (i = 0; i < MaxHashEntries; i++) {
		ptr = topcpuBackbone [i].next;
		while (ptr != NULL) {
			processList [p++] = ptr;
			ptr = ptr -> next;
		}
	}
	qsort (processList, p, sizeof (TopcpuT *), compareTOPCPU);
}

static char procname [17];

static void
displayDeltaTOPCPU ()
{
	int		i;
	clock_t	grandTotal;
	int		percentage;

	fetchSortTOPCPU ();
	grandTotal = 0;
	for (i = 0; i < processCount; i++) {
		grandTotal += processList [i] -> delta;
	}
	for (i = 0; (i < processCount) && (i < numTopcpuLines); i++) {
		if (!processList [i] -> delta) {
			break;
		}
		term_printf (6 + i, 0, opta, "%05d%-16.16s", processList [i] -> pid, processList [i] -> name);
		term_clear (TERM_CLS_EOL);
		percentage = (int) ((float) processList [i] -> delta * 100. / (float) grandTotal);
		sprintf (procname, "%%3d%%%d.%ds", percentage / 2, percentage / 2);
		term_printf (6 + i, 24, opta, procname, percentage, lines);
		if (percentage & 1) {	/* odd, do one of those half-char things... */
			term_printf (6 + i, 28 + percentage / 2, opta, "");
		}
		term_printf (6 + i, 77, opta, "");
		if (opth) {
			if (processList [i] -> prio > 5) {
				currentHistogramValue += percentage;
			}
		} 
	}
	term_printf (6 + i, 0, opta, "%s", "");
	if (opth) {
		for (; i < numTopcpuLines; i++) {	/* RK930828 was numTop - 1 */
			term_cur (7 + i, 0);
			term_clear (TERM_CLS_EOL);
		}
	} else {
		if (i < (numTopcpuLines - 1)) {
			term_cur (7 + i, 0);
			term_clear (TERM_CLS_EOS);
		}
	}
	term_flush ();
	free (processList);
}

static char buf [80];

/*
 *	displayHistogram ()
 *
 *	This routine is responsible for displaying the histogram on the bottom
 *	of the screen.
*/

static void
displayHistogram ()
{
	int		x;
	int		y;

	for (x = 0; x < 79; x++) {
		buf [x] = 0xdb;			/* fill with full bar */
	}
	buf [79] = 0;
	for (x = 0; x < 78; x++) {
		last80 [x] = last80 [x + 1];
	}
	last80 [78] = currentHistogramValue / 167L;
	for (y = 1; y < 30; y += 2) {
		for (x = 1; x < 79; x++) {
			if (last80 [x] == y) {
				buf [x] = 0xdc;				/* only get half a bar */
			} else if (last80 [x] < y) {
				buf [x] = ' ';
			}
		}
		term_printf (term_state.num_rows - (y / 2) - 2, 1, opta, "%s", buf + 1);
	}
	term_printf (term_state.num_rows - 2, 0, opta, "0");
	term_printf (term_state.num_rows - 16, 0, opta, "1");
	term_printf (term_state.num_rows - 9, 0, opta, "");
	histogramSamples++;
	histogramTotal += currentHistogramValue / 50L;
	term_printf (term_state.num_rows - 1, 0, opta, "Histogram sample %ld, average CPU utilization %2d%%", histogramSamples, (int) (histogramTotal / histogramSamples));
	term_flush ();
}

